/*
 * calcput - a command line utility to send a variable to
 * a TI89 Titanium calculator over the USB link
 *
 * Copyright (C) 2010 Michael Jordan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "ti_usb.h"
#include "ti_file.h"

int main(int argc, char *argv[]) {
	if(argc > 1) {
		string argument;
		argument.append(argv[1]);
		bool help = false;
		bool valid = false;
		if((argument == "-h") || (argument == "--help")) {
			valid = true;
			help = true;
		}
		if(!valid) help = true;
		if(help) {
			cout <<
					"\ncalcput is a utility to send individual files to a\n" <<
					"TI89 Titanium calculator over the USB link.\n\n" << flush;
			cout <<
					"After verifying that the calculator is present and it can\n" <<
					"be communicated with, the user enters a filename.  This\n" <<
					"utility will then present a list of any vars contained in\n" <<
					"the file and allow the user to choose which one to load.\n" <<
					flush;
			cout <<
					"Keep in mind that this is, in fact, free software, developed\n" <<
					"independently and therefore comes with no warranty.\n" << flush;
			return EX_OK;
		}
	}
	cout << "Searching for calculator..." << flush;
	ti_usb calc;

	if(calc.is_good()) {
		cout << "found calculator\n";
		string filename;
		cout << "Filename:  ";
		cin >> filename;
		ti_file file(filename);

		// Calling is_good() here makes sure that the file was opened successfully
		// and that the parser understood and was able to parse it.  Therefore
		// calls to file's members will now return valid data from the file.
		if(file.is_good()) {

			// Now we list the vars contained in the file to allow the user to
			// choose which one they would like sent to the calculator
			cout << right << setw(13) << "Folder:" <<
				left << setw(8) << "Name" <<
				right << setw(10) << "Size:" << " Archived Locked Type\tChecksum OK" <<
				endl << endl;
			int num_entries = file.num_entries();
			for(int i = 0; i < num_entries; i++) {
				cout << right << setw(4) << (i + 1) <<
						right << setw(8) << file.folder(i) << ":" <<
						left << setw(8) << file.name(i) <<
						right << setw(10) << file.data_size(i) << " " <<
						boolalpha << left << setw(8) << file.archived(i) <<
						" " << file.locked(i) << "\t" <<
						file.typeid_to_string(file.type_id(i)) << "\t" <<
						file.cs_ok(i) << endl;
			}
			cout << endl;
			cout << "Enter an item number (0 for all, -1 to quit):  ";
			int ent_no;
			cin >> ent_no;
			ent_no--;
			int start, end;
			if((ent_no >= -2) && (ent_no < num_entries)) {
				if(ent_no == -1) {
					start = 0;
					end = file.num_entries();
				} else {
					start = ent_no;
					end = ent_no + 1;
				}

				// Now we get a listing from the calculator.  We use this to
				// check to see if the var(s) the user wishes to send to the
				// calculator are already installed and thus do not need to be
				// sent.  We use this cached copy for our comparisons.  A new
				// listing could be obtained after sending each var, but in the
				// case of multiple vars, this would slow the process greatly.
				vector<ti_usb::ls_item_info> ls_items;
				cout << "Getting listing\n" << flush;
				calc.ls(ls_items);

				// Get the free ram and the free flash from the calculator.  We
				// use a cached copy of these values to track the free memory
				// on the calculator as we install multiple vars.  This check
				// could be done after sending each var, but it would make the
				// process considerably slower.
				int free_ram = calc.get_free_ram();
				int free_flash = calc.get_free_flash();
				for(int c = start; c < end; c++) {
					cout << "Checking to see if the var " << file.name(c) <<
							" is already on the calculator..." << flush;
					bool found_match = false;
					for(int i = 0; i < (int)ls_items.size(); i++)
						if((file.folder(c) == ls_items[i].parent_folder) &&
								(file.name(c) == ls_items[i].name))
									found_match = true;
					if(!found_match) {

						// We have determined that the var in question is not
						// listed as being installed on the calculator
						cout << "no - continuing...\n";
						bool enough_room;
						cout << "Checking for adequate storage on the calculator..." << flush;
						bool file_dest_flash;
						file_dest_flash = (file.is_flash() || file.archived(c));
						if(file_dest_flash) enough_room = (file.data_size(c) < free_flash);
						else enough_room = (file.data_size(c) < free_ram);
						if(enough_room) {

							// We have determined that there is enough storage
							// on the calculator to accept the var.  Now we pull
							// the actual data for the var out of the ti_file
							// object, format a header to accompany it, and send
							// the information to put_var().
							cout << "yes - continuing...\n";
							ti_usb::ls_item_info item;
							item.parent_folder = file.folder(c);
							item.name = file.name(c);
							item.size = file.data_size(c);
							item.type = file.type_id(c);
							item.archived = file.archived(c);
							item.locked = file.locked(c);
							unsigned char *data = new unsigned char[item.size];
							file.data(c, data);
							cout << "Sending item to the calculator..." << flush;
							int ret;
							ret = calc.put_var(item, data);
							if(ret == 0) {

								// If put_var() reports that the operation was
								// successful (with a 0 return value) we now
								// decrement the size of this var from the
								// appropriate cached storage value and continue.
								cout << "Success\n";
								if(file_dest_flash) free_flash -= file.data_size(c);
								else free_ram -= file.data_size(c);
							} else cout << "Failed\n";

							// Be sure and clean up - whether the var was
							// transmitted successfully or not.
							delete[] data;
						} else {
							cout << "no - skipping\n";
						}
					} else {
						cout << "yes - skipping\n";
					}
				}
			}
		} else {
			cout << "Invalid file - aborting\n";
			return -1;
		}
		return 0;
	}
	cout << "No calculator found on the USB bus\n";
	return -1;
}
