/*
 * ti_usb - a class used to handle USB comms for ti calculators
 *
 * Copyright (C) 2010 Michael Jordan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef TI_USB_H_
#define TI_USB_H_

#include <asm-generic/errno-base.h>
#include <errno.h>
#include <sysexits.h>
#include <libusb-1.0/libusb.h>
#include <sys/stat.h>
#include <dirent.h>
#include <cmath>
#include <ctime>
#include <iomanip>
#include <vector>
#include <fstream>
#include <sstream>
#include <iostream>
using namespace std;

// The ti_usb class handles the comms with the calculator

class ti_usb {
public:

	// The date_time class is used to transport the date and time values

	class date_time {
	public:
		int year;
		int month;
		int day;
		int hour;
		int minute;
		int second;

		date_time() {
			year = 0;
			month = 0;
			day = 0;
			hour = 0;
			minute = 0;
			second = 0;
		}
	};

	// When getting a listing of files on the calculator it is necessary
	// to declare a vector of ls_item_info objects to return the information

	class ls_item_info {
	public:
		string parent_folder;
		string name;
		int size;
		unsigned int type;
		bool archived;
		char app_var[4];
		bool locked;

		ls_item_info() {
			parent_folder = "";
			name = "";
			size = -1;
			type = 0;
			archived = false;
			app_var[0] = 0;
			app_var[1] = 0;
			app_var[2] = 0;
			app_var[3] = 0;
			locked = false;
		}
	};

	// The ti_usb constructor takes care of initializing libusb and then
	// searching for the calculator.  It also performs an initial ping in order
	// to verify that communications are functional.  After these functions are
	// completed successfully, an internal bit, is_inited is set to indicate
	// that things are now ready to be used.  Until this internal bit is set,
	// all other members will return error values when called.

	ti_usb();

	// The ti_usb destructor takes care of properly de-initializing libusb and
	// setting is_inited to false.

	~ti_usb();

	// Call the is_good() function to verify that initialization has been
	// completed successfully.  Until this function returns true, all of the
	// other functions will return error values.

	bool is_good();

	// Call set_homescreen() to cause the calculator to return to the Apps (home)
	// screen
	// Returns true on success - false on failure

	bool set_homescreen();

	// Call ls(vector<ti_usb::ls_item_info> &items) to get a listing of the files
	// on the calculator.  The listing is returned in a vector of ls_item_info
	// objects which can be iterated through to obtain the information that is
	// needed.
	// Returns true on success - false on failure

	bool ls(vector<ti_usb::ls_item_info> &items);

	// Call get_id_list() to get a list of the certificates stored on the
	// calculator.

	string get_id_list();

	// Call get_dbus_dev_type() to get the dbus device type number

	int get_dbus_dev_type();

	// Call get_id_string() to get the id string for the calculator
	// "TI-89 Titanium" for the Titanium

	string get_id_string();

	// Call get_free_ram() to find the amount of RAM that is currently
	// available to the user.

	int get_free_ram();

	// Call get_free_flash() to find the amount of flash memory that is
	// currently available to the user.

	int get_free_flash();

	// Call get_phys_flash() to find the physical amount of flash memory
	// installed on the calculator.

	int get_phys_flash();

	// Call get_phys_ram() to find the physical amount of RAM installed on the
	// calculator.

	int get_phys_ram();

	// Call get_prod_num() to return the product number for this calculator

	int get_prod_num();

	// Call get_hdw_version() to get the hardware version of the calculator

	int get_hdw_version();

	// Call get_boot_version() to find the boot version of the calculator

	string get_boot_version();

	// Call get_os_version() to return the version of the OS presently running
	// on the calculator

	string get_os_version();

	// Call to is_os_loaded() returns true if the OS is loaded on the calculator.
	// It returns false if it is not.

	bool is_os_loaded();

	// Call get_lcd_height() to find the height of the LCD display in pixels

	int get_lcd_height();

	// Call get_lcd_width() to find the width of the LCD display in pixels

	int get_lcd_width();

	bool disable_clock();

	bool enable_clock();

	// Call to is_clock_enabled() returns true if the hardware clock on the
	// calculator is enabled.  It returns false if it is not.

	bool is_clock_enabled();

	// Call set_date_fmt(int fmt) to set the date format on the calculator:
	// 0 = Y/M/D
	// 1 = M/D/Y
	// 2 = D/M/Y
	// 3 = M.D.Y
	// 4 = D.M.Y
	// 5 = Y.M.D
	// 6 = M-D-Y
	// 7 = D-M-Y
	// 8 = Y-M-D

	bool set_date_fmt(int fmt);

	// Call get_date_fmt() to return the date format currently set on the
	// calculator

	int get_date_fmt();

	// Call set_clock_mode(int mode) to set the clock mode - 0 for 12 hour and 1
	// for 24 hour

	bool set_clock_mode(int mode);

	// Call get_clock_mode() to return the clock mode - 0 for 12 hour and 1 for
	// 24 hour

	int get_clock_mode();

	// Call is_batt_good() to return the current state of the battery - true
	// for the battery is OK

	bool is_batt_good();

	// Call get_homescreen_state() to find out if the calculator is currently
	// displaying the home screen
	//
	// Returns true if the homescreen is displaying - false if not or on failure

	bool get_homescreen_state();

	// Call is_screen_split() to determine whether the screen is currently
	// displaying in a split state - returns true or false

	bool is_screen_split();

	// Call get_cert() to return the certificate for this calculator in a
	// string

	string get_cert();

	// Call get_lang_id() to return the language ID for the calculator - an
	// integer

	int get_lang_id();

	// Call get_sublang_id() to return the sub-language ID for the calculator
	// - an integer

	int get_sublang_id();

	// Call get_screenshot(string filename) to get a screenshot from the
	// calculator and store it in filename.  This member will not overwrite
	// filename if it already exists
	// Returns true on success - false on failure

	bool get_screenshot(string filename);

	// Call get_sys_time(date_time *dt) to get the date and time from the
	// PC.  The values are returned in the date_time object called dt.  This
	// date_time object must be created in the application and then its address
	// must be passed to this function as a pointer.
	// Returns true on success - false on failure

	bool get_sys_time(date_time *dt);

	// Call set_clock(date_time dt) to set the date and time stored in the
	// calculator's hardware clock.  The values are sent in a date_time object
	// created in the application.
	// Returns true on success - false on failure

	bool set_clock(date_time dt);

	// Call get_clock(date_time *dt) to get the date and time stored in the
	// calculator's hardware clock.  The values are returned in the date_time
	// object called dt.  This date_time object must be created in the application
	// and then its address must be passed to this function as a pointer.
	// Returns true on success - false on failure

	bool get_clock(date_time *dt);

	// Call delete_var(ls_item_info ls_item) to attempt to delete a var from the
	// calculator.  Some vars cannot be deleted this way.  So the return value
	// should be checked for a non-zero value to indicate that the operation was
	// not successful.

	int delete_var(ls_item_info ls_item);

	// Call get_var_date(ls_item_info ls_item, unsigned char *buf) to get the
	// actual var data (as opposed to just the header - which you get when you
	// request a directory listing).  ls_item describes the var which you wish
	// to retrieve.  buf is a pointer to an unsigned char array which will be
	// filled by this member function.  buf must be allocated in the application
	// then it must be deleted after it is no longer needed by the application.
	// Returns 0 on success - non-zero on failure

	int get_var_data(ls_item_info ls_item, unsigned char *buf);

	// Call var_save(ls_item_info ls_item, unsigned char *buf, string path) to
	// save a var once it has been retrieved from the calculator.  ls_item
	// contains the header information for the var.  buf is a pre-allocated
	// unsigned char array containing the actual data for the var.  path is a
	// string containing the actual path to save the file to.  path can be NULL
	// or it can be any valid path specifier ("./", "/usr/local/", etc.).  This
	// entry is not checked for a trailing slash and is pre-pended directly
	// onto the filename.
	// Returns 0 on success - non-zero on failure

	int var_save(ls_item_info ls_item, unsigned char *buf, string path);

	// typeid_to_string(int type_id) returns a string matching type_id as displayed
	// on the calculator's file listing (i.e. PRGM, FIG, STR, etc.)
	// Returns a string containing the information or "Unknown" if the type_id is
	// unknown.

	string typeid_to_string(int type_id);

	// A call to put_var(ls_item_info ls_item, unsigned char *data) will attempt
	// to send said variable to the calculator.  ls_item contains the header
	// information for the variable - size, folder, name, etc. and data is a
	// pointer to an unsigned char array containing the actual var.
	// Returns 0 on success - non-zero on failure

	int put_var(ls_item_info ls_item, unsigned char *data);

	// Call get_sys_directory(string dir, vector<std::string> &files) to get a
	// listing of all the files contained in the directory specified by dir.
	// The listing is returned in the vector files which must be created in
	// the application and disposed of in the application when no longer needed.
	// Returns 0 on success - non-zero on failure

	int get_sys_directory(string dir, vector<std::string> &files);

private:	/* ALL FUNCTIONS AFTER THIS POINT ARE PRIVATE - NOT TO BE USED
			 BY THE APPLICATION*/

	// The convert_jd_to_gd(float jd, int *month, int *day, int *year) member
	// converts a julian day value to gregorian calendar values(M, D, Y).  This is
	// necessary because the typical *nix system calculates time with a signed
	// 32 bit integer based on January 1, 1900 - whereas the calculator uses an
	// unsigned 32 bit integer based on January 1, 1997
	//
	// To overcome this incompatibility julian day time measurement is used as
	// an intermediate format

	void convert_jd_to_gd(float jd, int *month, int *day, int *year);

	// The to_julian_days(int Y, int M, int D) member accepts a gregorian date
	// (Y, M, D) and converts it to a julian day value.  This is necessary
	// because the typical *nix system calculates time with a signed  32 bit
	// integer based on January 1, 1900 - whereas the calculator uses an
	// unsigned 32 bit integer based on January 1, 1997
	//
	// To overcome this incompatibility julian day time measurement is used as
	// an intermediate format

	long double to_julian_days(int Y, int M, int D);

	// The get_secs_ti(date_time des_date, unsigned int *result) member takes
	// the date and time which you wish to have displayed on the calculator in
	// gregorian date format (in a date_time object) and returns the value
	// in result.  It returns true on success and false on failure.

	bool get_secs_ti(date_time des_date, unsigned int *result);

	bool is_inited;
	libusb_context *ctx;
	libusb_device_handle *hdl;
	enum direction {in = (1 | LIBUSB_ENDPOINT_IN), out = (2 | LIBUSB_ENDPOINT_OUT)};

	// The class packet is an internal class used in the transmitting of
	// information to and from the calculator

	class packet {
	public:

		packet() {
			timeout = 500;
			multi_in_progress = false;
			payload = NULL;
		}

		libusb_device_handle *hdl;
		int raw_len;
		int raw_type;
		int virt_len;
		int virt_type;
		int payload_len;
		int timeout;
		bool multi_in_progress;
		unsigned char *payload;
	};

	// The class param_req is an internal class used in the formation of
	// parameter request packets

	class param_req {
	public:
		int param_num;
		int data_len;
		unsigned char *data;
	};

	// The class transfer_info is an internal class used in the transmitting of
	// information to and from the calculator

	class transfer_info {
	public:
		enum direction direct;
		libusb_device_handle *hdl;
		unsigned char *data;
		int *len;
		int timeout;
		int actual;
	};

	// The class ctl_transfer_info is an internal class used in the transmitting
	// of control packets to and from the calculator

	class ctl_transfer_info {
	public:
		libusb_device_handle *hdl;
		unsigned char bmRequestType;
		unsigned char bRequest;
		unsigned int wValue;
		unsigned int wIndex;
		unsigned int wLength;
		unsigned int timeout;
		unsigned char *data;
	};

	// flash_head is used when saving a flash var from the calculator to fill in
	// the required data and write the file to disk

	class flash_head {
	public:
		unsigned char version[2];
		unsigned char bcd_date[4];
		string name;
		int dev_type;
		int data_type;
		int hdw_id;
		int data_size;
		unsigned char *data;
	};

	// bulk_transfer(transfer_info *tinfo) accepts a transfer_info object pointer
	// and based on the information therein handles the information to or from
	// the calculator as needed
	// Returns true on success - false on failure

	bool bulk_transfer(transfer_info *tinfo);

	// The parse_ls_pkt(unsigned char *pkt, vector<ti_usb::ls_item_info> &res)
	// member is a private member used in getting a directory listing from
	// the calculator.  It handles parsing the raw packets, adding new items
	// to the vector<ti_usb::ls_item_info> and populating the fields with the
	// parsed information

	int parse_ls_pkt(unsigned char *pkt, vector<ti_usb::ls_item_info> &res);

	// The parameter_request(param_req *req) member accepts a param_req pointer,
	// forms and executes the parameter request and then returns the data in the
	// param_req object
	// Returns true on success - false on failure

	bool parameter_request(param_req *req);

	// The parameter_set(param_req req) member accepts a param_req object, forms
	// and executes the parameter set request
	// Returns true on success - false on failure

	bool parameter_set(param_req req);

	// The member send_pkt(packet pk) forms a packet based on the information
	// in the packet object and then sends the packet to the calculator via
	// bulk_transfer(transfer_info *tinfo)

	int send_pkt(packet pk);

	// The send_virt_pkt_ack() member forms and sends a virtual packet
	// acknowledgment to the calculator

	int send_virt_pkt_ack();

	// The find_calc() member iterates through the items on the USB bus after
	// libusb's initialization and searches for a calculator by matching VID:PID

	int find_calc();

	// If a kernel module is found handling the device we wish to use, the
	// unload_kern_driver() member attempts to unload it

	int unload_kern_driver();

	// The recv_pkt(packet *pkt_prsd) member receives a packet from the calculator
	// and breaks it down to populate pkt_prsd with (if applicable) raw length,
	// raw type, virtual length and virtual type.  Then it attaches the remaining
	// payload to the pkt_prsd object for use in the calling member.  In order
	// to prevent memory leaking it is critical to delete pkt_prsd->payload
	// after the call to this member and after its contents are no longer needed

	int recv_pkt(packet *pkt_prsd);

	// init_chat() handles the buffer size request, buffer size allocation, and
	// other miscellaneous packets needed to begin every transaction with the
	// calculator

	int init_chat();

	// control_transfer(ctl_transfer_info *cti) is used to handle the endpoint
	// resets which are used to keep the calculator communicating with us

	bool control_transfer(ctl_transfer_info *cti);

	// reset_pipe_long() forms the long endpoint reset sequence

	bool reset_pipe_long();

	// reset_pipe_short() forms the short endpoint reset sequence

	bool reset_pipe_short();

	// write_file(string filename, unsigned char *buf, int size) is called in
	// order to save a var to disk once it is retrieved from the calculator.
	// filename contains the desired path/filename.  buf is an unsigned char
	// array which contains the actual var data.  size is the size (in bytes) of
	// buf.
	// Returns true on success - false on failure

	bool write_file(string filename, unsigned char *buf, int size);

	// typeid_to_filext(int type_id) accepts a type ID and returns the
	// corresponding file extension to save the var.
	// Returns a NULL string if the type ID is unknown.

	string typeid_to_filext(int type_id);

};

#endif /* TI_USB_H_ */
